//+------------------------------------------------------------------+
//|                                                           EA.mqh |
//|                                                Maxim A.Kuznetsov |
//|                                                       nektomk.ru |
//+------------------------------------------------------------------+
#property copyright "Maxim A.Kuznetsov"
#property link      "nektomk.ru"
#property version   "1.00"
#property strict

#include <stdlib.mqh>

class Ind {
public:
   Ind(string path,string symbol="",int tf=0);
   ~Ind();
   virtual int Total();
   virtual double Value(int mode,int bar=0);
   virtual double Sig(int mode,int bar=0);
public:
   string symbol;
   int tf;
   string path;
   int maxAttempts;  // maximum attempts for data retr.
   int msOnError;    // sleep when retr. error reached
};
Ind::Ind(string _path,string _symbol,int _tf) {
   path=_path;
   if (_symbol=="") symbol=_Symbol;
   else symbol=_symbol;
   if (_tf==0) tf=_Period;
   else tf=_tf;
   maxAttempts=3;
   msOnError=100;
};
Ind::~Ind() {
}
int Ind::Total() {
   return iBars(symbol,tf);
}
double Ind::Value(int mode,int bar) {
   return EMPTY_VALUE;
}
enum PRIMA_IND_BUFFER {
   PRIMA_IND_MA1,
   PRIMA_IND_MA2,
   PRIMA_IND_AO_UP,
   PRIMA_IND_AO_UP2,
   PRIMA_IND_AO_DOWN,
   PRIMA_IND_AO_DONW2,
   PRIMA_IND_BUY,
   PRIMA_IND_SELL
};
class PrimaInd: public Ind {
public:
   PrimaInd(int period1,int period2,int distance1,int distance2,string path="market\\prima",string symbol="",int tf=0);
   ~PrimaInd();
   bool IsCorrect();
   double Value(int mode,int bar=0);
   double Sig(int mode,int bar=0);
   double SigBuy(int bar=0);
   double SigSell(int bar=0);
   int Op(int bar);
   int Prev(int mode=-1,int bar=0);
   int PrevSame(int mode,int bar=0);
   int PrevContr(int mode,int bar=0);
   int Next(int mode,int bar=INT_MAX);
   int NextSame(int mode,int bar=INT_MAX);
   int NextContr(int mode,int bar=INT_MAX);
public:   
   int period1;
   int period2;
   int distance1;
   int distance2;
};
PrimaInd::PrimaInd(int _period1,int _period2,int _distance1,int _distance2,string _path,string _symbol,int _tf):Ind(_path,_symbol,_tf) 
{
   period1=_period1;
   period2=_period2;
   distance1=_distance1;
   distance2=_distance2;
}
PrimaInd::~PrimaInd()
{
}
bool PrimaInd::IsCorrect(void) {
   if (period1<2 || period2<2 || period1==period1)
      return false;
   if (maxAttempts<=0) maxAttempts=1;   
   return true;
}
double PrimaInd::Value(int mode,int bar=0) {
   return iCustom(symbol,tf,path,
      period1,period2,
      distance1,distance2,
      false,false,false,   // no graphics
      false,false,false,   // no alerts
      mode,bar);
}
double PrimaInd::Sig(int mode,int bar=0) {
   double value;
   int attempt;
   int err;
   value=0.0;
   for(attempt=0;attempt<maxAttempts;attempt++) {
      ResetLastError();
      value=Value(mode,bar);
      if (value==EMPTY_VALUE) {
         err=GetLastError();
         if (err!=ERR_NO_ERROR) {
            PrintFormat("Error on Sig():%d %s",err,ErrorDescription(err) );
            RefreshRates();
            if (attempt) Sleep(msOnError);
            continue;
         } else break;
      }   
   }
   return value;
}
double PrimaInd::SigBuy(int bar) {
   return Sig(PRIMA_IND_BUY,bar);
}
double PrimaInd::SigSell(int bar) {
   return Sig(PRIMA_IND_SELL,bar);
}
int PrimaInd::Op(int bar) {
   if (SigBuy(bar)!=EMPTY_VALUE) return OP_BUY;
   if (SigSell(bar)!=EMPTY_VALUE) return OP_SELL;
   return -1;
}
int PrimaInd::Prev(int mode,int bar) {
   int pos,total;
   if (bar<0) return -1;
   total=Total();
   for(pos=bar;pos>total;pos++) {
      if (mode==-1) {
         if (SigBuy(pos)!=EMPTY_VALUE || SigSell(pos)!=EMPTY_VALUE) {
            return pos;
         }
      } else {
         if (Sig(mode,pos)!=EMPTY_VALUE) return pos;
      }
   }
   return -1;
}
int PrimaInd::PrevContr(int mode,int bar) {
   if (bar<0 || bar==INT_MAX) return bar;
   if (mode==-1) {
      mode=Prev(mode,bar);
      if (mode==-1) return -1;
   }
   if (mode==PRIMA_IND_BUY) return Prev(PRIMA_IND_SELL,bar);
   return Prev(PRIMA_IND_BUY,bar);
}
int PrimaInd::Next(int mode,int bar) {
   int pos,total;
   if (bar<0) return bar;
   if (bar==INT_MAX) bar=Total()-1;
   total=Total()-1;
   for(pos=total;pos<=0;pos--) {
      if (mode==-1) {
         if (SigBuy(pos) || SigSell(pos)) return pos;
      } else {
         if (Sig(mode,pos)!=EMPTY_VALUE) return pos;
      }
   }
   return -1;      
}
int PrimaInd::NextContr(int mode,int bar) {
   if (bar<0 || bar==INT_MAX) return bar;
   if (mode==-1) {
      mode=Op(Next(-1,bar));
      if (mode==-1) return -1;
   }   
   if (mode==PRIMA_IND_BUY) return Next(PRIMA_IND_SELL,bar);
   return Next(PRIMA_IND_BUY,bar);
}


bool Button(string name,string title,color clr,int width,int height,int x,int y) {
   if (!ObjectCreate(0,name,OBJ_BUTTON,0,0,0) && ObjectType(name)!=OBJ_BUTTON) return false;
   ObjectSetInteger(0,name,OBJPROP_ANCHOR,ANCHOR_LEFT_UPPER);
   ObjectSetInteger(0,name,OBJPROP_CORNER,CORNER_RIGHT_UPPER);
   ObjectSetInteger(0,name,OBJPROP_COLOR,clr);
   ObjectSetInteger(0,name,OBJPROP_XSIZE,width);
   ObjectSetInteger(0,name,OBJPROP_YSIZE,height);
   ObjectSetInteger(0,name,OBJPROP_XDISTANCE,x);
   ObjectSetInteger(0,name,OBJPROP_YDISTANCE,y);
   ObjectSetInteger(0,name,OBJPROP_STATE,0);
   ObjectSetString(0,name,OBJPROP_FONT,"Arial");
   ObjectSetInteger(0,name,OBJPROP_FONTSIZE,8);
   ObjectSetString(0,name,OBJPROP_TEXT,title);
   ObjectSetInteger(0,name,OBJPROP_HIDDEN,true);
   return true;
}
bool Label(string name,string title,color clr,int width,int height,int x,int y) {
   if (!ObjectCreate(0,name,OBJ_LABEL,0,0,0) && ObjectType(name)!=OBJ_LABEL) return false;
   ObjectSetInteger(0,name,OBJPROP_ANCHOR,ANCHOR_LEFT_UPPER);
   ObjectSetInteger(0,name,OBJPROP_CORNER,CORNER_RIGHT_UPPER);
   ObjectSetInteger(0,name,OBJPROP_COLOR,clr);
   ObjectSetInteger(0,name,OBJPROP_XSIZE,width);
   ObjectSetInteger(0,name,OBJPROP_YSIZE,height);
   ObjectSetInteger(0,name,OBJPROP_XDISTANCE,x);
   ObjectSetInteger(0,name,OBJPROP_YDISTANCE,y);
   ObjectSetString(0,name,OBJPROP_FONT,"Arial");
   ObjectSetInteger(0,name,OBJPROP_FONTSIZE,8);
   ObjectSetString(0,name,OBJPROP_TEXT,title);
   ObjectSetInteger(0,name,OBJPROP_HIDDEN,true);
   return true;
}
enum PRIMA_EVENTS {
   PRIMA_EV_BUY,
   PRIMA_EV_BUY_DONE,
   PRIMA_EV_SELL,
   PRIMA_EV_SELL_DONE,
   PRIMA_EV_CLOSEALL,
   PRIMA_EV_CLOSEALL_DONE,
   PRIMA_EV_CLOSELAST,
   PRIMA_EV_CLOSELAST_DONE
};
enum ENUM_REQUEST {
   REQUEST_NONE,
   REQUEST_BUY,
   REQUEST_SELL,
   REQUEST_CLOSE_LAST,
   REQUEST_CLOSE_ALL
   
};
string modeTitle[]={
   "Auto",
   "Delay",
   "Manual"
};
class EA;
class Panel {
public:
   Panel(EA *_ea,string _prefix="pan");
   ~Panel();
   void Show();
   void Hide();
   virtual bool OnClick(const long lparam,const double dparam,const string &sparam);
   void OnClickBuy(string button);
   void OnClickSell(string button);
   void OnClickCloseAll(string button);
   void OnClickCloseLast(string button);
   void BeginRequest(ENUM_REQUEST _request,datetime _requestTimeout,bool _accept=true);
   void EndRequest(ENUM_REQUEST _request);
   void BeginBuy();void EndBuy();
   void BeginSell();void EndSell();
   void BeginCloseAll();void EndCloseAll();
   void BeginCloseLast();void EndCloseLast();
   void UpdateTotals(double totalProfit,double lastProfit);
   virtual bool OnCustomEvent(const int ev,const long lparam,const double dparam,const string &sparam);
   virtual void OnTimer();
   void StateText(string text,color clr,int timeout);
public:
   EA *ea;
   string prefix;
   int mode;   //   
   //  
   string labelState;         //  - / 
   string buttonMode[3];      //    (-  , dealy- , manual-  
   string buttonAccept,buttonDecline;  // -/
   
   string buttonSell,buttonBuy;  // /
   string buttonCloseAll;        //   
   string buttonCloseLast;       //    
   string buttonCloseBuy,buttonCloseSell; //    buy,sell
   
   datetime timeToClearState; //        
};
Panel::Panel(EA *_ea,string _prefix) {
   int t;
   ea=_ea;
   prefix=_prefix;
   if (ea!=NULL) mode=ea.mode;
   else mode=0;
   ObjectsDeleteAll(0,prefix);
   for(t=0;t<3;t++) {
      buttonMode[t]=prefix+".mode"+IntegerToString(t);
   }
   labelState=prefix+".state";
   buttonAccept=prefix+".accept";
   buttonDecline=prefix+".decline";
   buttonSell=prefix+".sell";
   buttonBuy=prefix+".buy";
   buttonCloseAll=prefix+".closeAll";
   buttonCloseLast=prefix+".closeLast";
   timeToClearState=0;
}
Panel::~Panel() {
   ObjectsDeleteAll(0,prefix);
}
Panel::Show() {
   int t;
   int x,y;
   int bw=180; //  ..   6
   int screenWidth,screenHeight;
   x=60;
   y=20;
   screenWidth=(int)ChartGetInteger(0,CHART_WIDTH_IN_PIXELS);
   screenHeight=(int)ChartGetInteger(0,CHART_HEIGHT_IN_PIXELS);
   //   -    
   for(t=0;t<3;t++) {
      Button(buttonMode[t],modeTitle[t],(mode==t?clrOrange:clrGreen),bw/3,y,bw/3*(3-t),y);
      if (t==mode) {
         ObjectSetInteger(0,buttonMode[t],OBJPROP_COLOR,clrOrange);
         ObjectSetInteger(0,buttonMode[t],OBJPROP_STATE,1);
         ObjectSetString(0,buttonMode[t],OBJPROP_FONT,"Arial Black");
      }
   }   
   //  ()  
   Label(labelState,"State",clrGray,bw,20,bw,y+20);
   //  buy-sell|close|| close all|~all buy|~all sell
   //  , 
   y=screenHeight/2-80/2;
   x=bw*3;
   //  Accept/Decline
   Button(buttonAccept,"Accept",clrGray,bw,20,bw,y);
   //  buy/sell
   Button(buttonSell,"Sell",clrRed,bw/2,20,bw,y+20);
   Button(buttonBuy, "Buy",clrBlue,bw/2,20,bw-bw/2,y+20);
   //  -   
   Button(buttonCloseLast,"Close last",clrDarkGray,bw,20,bw,y+40);
   //  -  
   Button(buttonCloseAll,"Close all",clrDarkGray,bw,20,bw,y+70);
}

void Panel::StateText(string text,color clr,int timeout=0) {
   if (labelState=="") return;
   ObjectSetString(0,labelState,OBJPROP_TEXT,text);
   ObjectSetInteger(0,labelState,OBJPROP_COLOR,clr);
   if (timeout!=0)
      timeToClearState=TimeCurrent()+timeout;   
   else
      timeToClearState=0;
}
void Panel::OnTimer(void) {
   if (timeToClearState!=0 && labelState!="") {
      if (timeToClearState<TimeCurrent()) {
         ObjectSetString(0,labelState,OBJPROP_TEXT,"");
         ObjectSetInteger(0,labelState,OBJPROP_COLOR,clrGray);
      }
   }
}
bool Panel::OnClick(const long lparam,const double dparam,const string &sparam) {
   int t;
   for(t=0;t<3;t++) {
      if (sparam==buttonMode[t]) {
         ObjectSetInteger(0,buttonMode[mode],OBJPROP_STATE,0);
         ObjectSetInteger(0,buttonMode[t],OBJPROP_STATE,1);
         if (ea!=NULL) {
            ea.SetMode(t);
            mode=ea.mode;
         } else {
            mode=t;
         }
         return true;
      }
   }
  
   //      - ""   ,    
   if (sparam==buttonSell || sparam==buttonBuy || sparam==buttonCloseAll) {
      //if (ObjectGetInteger(0,sparam,OBJPROP_STATE)!=0) return true;  //   
      if (sparam==buttonBuy) {
         //EventChartCustom(ChartID(),PRIMA_EV_BUY,0,0,"");
         OnClickBuy(buttonBuy);
      }
      else if (sparam==buttonSell) {
         //EventChartCustom(ChartID(),PRIMA_EV_SELL,0,0,"");
         OnClickSell(buttonSell);
      }
      else if (sparam==buttonCloseAll) OnClickCloseAll(sparam);
      else if (sparam==buttonCloseLast) OnClickCloseLast(sparam);
      return true;
   }
   return false;
}
void Panel::OnClickBuy(string button) {
   if (ea!=NULL) ea.Buy();
   ObjectSetInteger(0,button,OBJPROP_STATE,0);
}
void Panel::OnClickSell(string button) {
   if (ea!=NULL) ea.Sell();
   ObjectSetInteger(0,button,OBJPROP_STATE,0);
}
void Panel::OnClickCloseAll(string button) {
   if (ea!=NULL) ea.CloseAll();
   ObjectSetInteger(0,button,OBJPROP_STATE,0);
}
void Panel::OnClickCloseLast(string button) {
   if (ea!=NULL) ea.CloseLast();
   ObjectSetInteger(0,button,OBJPROP_STATE,0);
}

void Panel::BeginBuy() {
   ObjectSetInteger(0,buttonBuy,OBJPROP_STATE,1);
}
void Panel::EndBuy() {
   ObjectSetInteger(0,buttonBuy,OBJPROP_STATE,0);
}
void Panel::BeginSell() {
   ObjectSetInteger(0,buttonSell,OBJPROP_STATE,1);
}
void Panel::EndSell() {
   ObjectSetInteger(0,buttonSell,OBJPROP_STATE,0);
}
void Panel::BeginCloseAll() {
   ObjectSetInteger(0,buttonCloseAll,OBJPROP_STATE,1);
}
void Panel::EndCloseAll() {
   ObjectSetInteger(0,buttonCloseAll,OBJPROP_STATE,0);
}
void Panel::BeginCloseLast() {
   ObjectSetInteger(0,buttonCloseLast,OBJPROP_STATE,1);
}
void Panel::EndCloseLast() {
   ObjectSetInteger(0,buttonCloseLast,OBJPROP_STATE,0);
}
void Panel::UpdateTotals(double totalProfit,double lastProfit) {
   if (totalProfit==0.0) {
      ObjectSetString(0,buttonCloseAll,OBJPROP_TEXT,"Close All");
      ObjectSetInteger(0,buttonCloseAll,OBJPROP_COLOR,clrGray);
   } else {
      ObjectSetString(0,buttonCloseAll,OBJPROP_TEXT,"Close All "+(totalProfit>0?"+":"")+DoubleToString(totalProfit,2));
      if (totalProfit>0) {
         ObjectSetInteger(0,buttonCloseAll,OBJPROP_COLOR,clrDarkGreen);
      } else {
         ObjectSetInteger(0,buttonCloseAll,OBJPROP_COLOR,clrDarkOrange);
      }
   }
   if (lastProfit==0.0) {
      ObjectSetString(0,buttonCloseLast,OBJPROP_TEXT,"Close Last");
      ObjectSetInteger(0,buttonCloseLast,OBJPROP_COLOR,clrGray);
   } else {
      ObjectSetString(0,buttonCloseLast,OBJPROP_TEXT,"Close Last "+(lastProfit>0?"+":"")+DoubleToString(lastProfit,2));
      if (lastProfit>0) {
         ObjectSetInteger(0,buttonCloseLast,OBJPROP_COLOR,clrDarkGreen);
      } else {
         ObjectSetInteger(0,buttonCloseLast,OBJPROP_COLOR,clrDarkOrange);
      }
   }
}
void Panel::BeginRequest(ENUM_REQUEST _request,datetime _requestTimeout,bool _accept) {
   switch(_request) {
      case REQUEST_BUY:
         ObjectSetInteger(0,buttonBuy,OBJPROP_BGCOLOR,clrBlue);ObjectSetInteger(0,buttonBuy,OBJPROP_COLOR,clrWhite);
         if (mode==1 && _accept) {
            ObjectSetString(0,buttonBuy,OBJPROP_TEXT,"Buy "+IntegerToString(_requestTimeout-TimeCurrent()) );
            ObjectSetString(0,buttonAccept,OBJPROP_TEXT,"Don`t BUY");
         } else {
            ObjectSetString(0,buttonBuy,OBJPROP_TEXT,"Buy");
            ObjectSetString(0,buttonAccept,OBJPROP_TEXT,"Don`t BUY "+IntegerToString(_requestTimeout-TimeCurrent()) );
         }
         break;
      case REQUEST_SELL:
         ObjectSetInteger(0,buttonSell,OBJPROP_BGCOLOR,clrRed);ObjectSetInteger(0,buttonSell,OBJPROP_COLOR,clrWhite);
         if (mode==1 && _accept ) {
            ObjectSetString(0,buttonSell,OBJPROP_TEXT,"Sell "+IntegerToString(_requestTimeout-TimeCurrent()) );
            ObjectSetString(0,buttonAccept,OBJPROP_TEXT,"Don`t SELL");
         } else {
            ObjectSetString(0,buttonSell,OBJPROP_TEXT,"Sell" );
            ObjectSetString(0,buttonAccept,OBJPROP_TEXT,"Don`t SELL "+IntegerToString(_requestTimeout-TimeCurrent()) );
         }
         break;
   }
}

void Panel::EndRequest(ENUM_REQUEST _request) {
   ObjectSetString(0,buttonAccept,OBJPROP_TEXT,"");

   ObjectSetInteger(0,buttonBuy,OBJPROP_BGCOLOR,clrLightGray);
   ObjectSetInteger(0,buttonBuy,OBJPROP_COLOR,clrBlue);
   ObjectSetString(0,buttonBuy,OBJPROP_TEXT,"Buy");

   ObjectSetInteger(0,buttonSell,OBJPROP_BGCOLOR,clrLightGray);
   ObjectSetInteger(0,buttonSell,OBJPROP_COLOR,clrRed);
   ObjectSetString(0,buttonSell,OBJPROP_TEXT,"Sell");
}
bool Panel::OnCustomEvent(const int ev,const long lparam,const double dparam,const string &sparam) {
   switch(ev) {
      case PRIMA_EV_BUY_DONE: ObjectSetInteger(0,buttonBuy,OBJPROP_STATE,0); return true;
      case PRIMA_EV_SELL_DONE: ObjectSetInteger(0,buttonSell,OBJPROP_STATE,0); return true;
      case PRIMA_EV_CLOSEALL_DONE: ObjectSetInteger(0,buttonCloseAll,OBJPROP_STATE,0); return true;
      case PRIMA_EV_CLOSELAST_DONE: ObjectSetInteger(0,buttonCloseLast,OBJPROP_STATE,0); return true;
   }
   return false;
}

Panel::Hide() {
   ObjectsDeleteAll(0,prefix);
}
class EA {
public:
   EA(ENUM_TIMEFRAMES tf,string name,int magik);
   ~EA();
   virtual int Setup();
   virtual bool IsCorrect();
   virtual void OnConnect();
   virtual void OnDisconnect();
   virtual void OnTimer();
   virtual void EveryTick();
   virtual void EveryBar();
   virtual bool OnEvent( const int id, const long& lparam,  const double& dparam, const string& sparam);
   virtual void OnError(string msg);
   virtual bool AllowedOpen(int op);
   virtual bool AllowedClose();
   void Update();
   void SetMode(int _mode);
   // trade functions
   void RequestBuy();
   void RequestSell();
   void RequestCloseAll();
   void RequestCloseLast();
   int TradeByMarket(int op);
   virtual int Buy();
   virtual int Sell();
   bool CalcOrder(int op,double &lots,double &stopLoss,double &takeProfit,datetime &expire);
   double CloseAll(int op=-1);
   double CloseLast(int op=-1);
   double CloseExpired(int seconds,int op=-1);
   double CloseSelected();
   
   void DrawSignal(int op,datetime time,double price);
public:
   ENUM_REQUEST request;   //  
   datetime requestTimeout;
   int mode;   //   0- 1-  2- 
   string modeVarName;  //     
   int secToConfirm;    //    
   Panel *panel;
   string name;
   int magik;
   int tf;
   datetime time0;
   datetime now;
   double lots;
   int stopLoss;
   int takeProfit;
   int ttl;
   int slippage;
   bool connected;   //  - /   
   
   //   
   //   
   //       -  
   //        -    secToConfirm
   //        -   
   int weekdayFrom;  //   
   int weekdays;     // -  
   int hourFrom;     //   
   int hours;        // -  
   bool leaveMarket;  //     ?
   void SetTradeTime(int _weekdayFrom,int _weekdays,int _hourFrom,int _hours,bool _leave);
   bool IsTradeTime(datetime time=0);
};

EA::EA(ENUM_TIMEFRAMES _tf,string _name,int _magik)
{
   if (!IsTesting()) ObjectsDeleteAll(0,name);
   request=REQUEST_NONE;
   requestTimeout=0;
   secToConfirm=90;
   if (_tf==0) tf=_Period;
   else tf=_tf;
   name=name;
   magik=_magik;
   time0=0;
   now=0;
   lots=MarketInfo(_Symbol,MODE_MINLOT);
   stopLoss=0.0;
   takeProfit=0.0;
   ttl=0;
   weekdayFrom=0;
   weekdays=7;
   hours=0;
   hours=24;
   leaveMarket=false;
   if (!IsTesting() && !IsOptimization()) {
      modeVarName="name"+_Symbol+IntegerToString(magik)+"_mode";
      if (GlobalVariableCheck(modeVarName)) {
         mode=(int)GlobalVariableGet(modeVarName);
      }
      panel=new Panel(GetPointer(this),"prima");
      panel.Show();
      Update();
   } else {
      mode=0;
      panel=NULL;
   }
   if (!IsTesting() && !IsOptimization()) {
      connected=IsConnected();
      if (connected) OnConnect();
   } else {
      connected=true;
   }
}
EA::~EA() {
   if (panel!=NULL) delete panel;
   if (!IsTesting()) ObjectsDeleteAll(0,name);
}
void EA::SetTradeTime(int _weekdayFrom,int _weekdays,int _hourFrom,int _hours,bool _leave) {
   weekdayFrom=_weekdayFrom;
   weekdays=_weekdays;
   hourFrom=_hourFrom;
   hours=_hours;
   leaveMarket=_leave;
}
bool EA::IsTradeTime(datetime time=0) {
   int d,h,weekdayTo,hourTo;
   if (time==0) time=TimeCurrent();
   if (weekdays>=7 && hours>=24) return true;
   d=TimeDayOfWeek(time);
   h=TimeHour(time);
   weekdayTo=(weekdayFrom+weekdays)%7;
   hourTo=(hourFrom+hours)%24;
   if (weekdayFrom<weekdayTo) {
      if (d<weekdayFrom || d>weekdayTo) return false;
   } else {
      if (d>weekdayTo && d<weekdayFrom) return false;
   }
   if (hourFrom<hourTo) {
      if (h<hourFrom || h>hourTo) return false;
   } else {
      if (h>hourTo && h<hourFrom) return false;
   }
   return true;
}

void EA::SetMode(int _mode) {
   mode=_mode;
   if (!IsTesting() && !IsOptimization()) {
      GlobalVariableSet(modeVarName,mode);
   }
   if (panel!=NULL) {
      if (mode==0) panel.StateText("AutoTrade mode",clrGreen,30);
      else if (mode==1) panel.StateText("Confirmation mode",clrGreen,30);
      else panel.StateText("Manual trading",clrGreen,30);
   }
}
void EA::OnConnect(void) {
   if (panel!=NULL) {
      panel.StateText("Connected",clrGray,30);
      Update();
   }
}
void EA::OnDisconnect(void) {
   if (panel!=NULL) panel.StateText("Disconnected",clrDarkOrange);
}
void EA::EveryTick() {
   now=TimeCurrent();
   if (time0!=iTime(_Symbol,tf,0)) {
      if (!IsTesting() && !IsOptimization()) {
         //       -  
         if (time0==0) {
            time0=iTime(_Symbol,tf,0);
            OnConnect();
         } else {
            EveryBar();
         }
      } else {
         //       
         EveryBar();
      }
   }
}
void EA::EveryBar() {
   time0=iTime(_Symbol,tf,0);
}
void EA::OnError(string msg) {
   int err;
   err=GetLastError();
   Alert("ERROR "+msg+" : "+IntegerToString(err)+" "+ErrorDescription(err));
}
void EA::OnTimer() {
   bool con;
   if (!IsTesting() && !IsOptimization()) {
      con=IsConnected();
      if (con!=connected) {
         connected=con;
         if (!connected) OnDisconnect();
         else OnConnect();
      }
   }
   if (panel!=NULL) {
      Update();
      panel.OnTimer();
   }
   if (mode!=0 && request!=REQUEST_NONE) {
      if (TimeCurrent()>=requestTimeout) {
         if (mode==1) {
            switch(request) {
               case REQUEST_BUY: Buy(); break;
               case REQUEST_SELL: Sell(); break;
               case REQUEST_CLOSE_ALL: CloseAll(); break;
               case REQUEST_CLOSE_LAST: CloseLast(); break;
            }
         }
         if (panel!=NULL) panel.EndRequest(request);
      }
   }
}
bool EA::OnEvent( const int id, const long& lparam,  const double& dparam, const string& sparam) {
   if (panel!=NULL) {
      if (id==CHARTEVENT_OBJECT_CLICK && panel.OnClick(lparam,dparam,sparam) ) return true;
   }
   if (id>=CHARTEVENT_CUSTOM && id<=CHARTEVENT_CUSTOM_LAST) {
      if (panel!=NULL && panel.OnCustomEvent(id-CHARTEVENT_CUSTOM,lparam,dparam,sparam) ) return true;
      switch(id-CHARTEVENT_CUSTOM) {
         case PRIMA_EV_BUY: 
            Buy();
            EventChartCustom(0,PRIMA_EV_BUY_DONE);
            return true;
         case PRIMA_EV_SELL:
            Sell();
            EventChartCustom(0,PRIMA_EV_SELL_DONE);
            return true;
         case PRIMA_EV_CLOSEALL:
            CloseAll();
            EventChartCustom(0,PRIMA_EV_CLOSEALL_DONE);
            return true;
         case PRIMA_EV_CLOSELAST:
            CloseLast();
            EventChartCustom(0,PRIMA_EV_CLOSELAST_DONE);
            return true;
      }
   }
   return false;
}

bool EA::AllowedOpen(int op) {
   if (!IsTesting() && !IsOptimization()) {
      if (AccountInfoInteger(ACCOUNT_TRADE_EXPERT)==0 || MarketInfo(_Symbol,MODE_TRADEALLOWED)==0) {
         return false;
      }
      if (!connected) return false;
   }
   return true;
}
bool EA::AllowedClose() {
   if (!IsTesting() && !IsOptimization()) {
      if (AccountInfoInteger(ACCOUNT_TRADE_EXPERT)==0 || MarketInfo(_Symbol,MODE_TRADEALLOWED)==0) {
         return false;
      }
      if (!connected) return false;
   }
   return true;
}
bool EA::IsCorrect(void) {
   return true;
}
bool EA::CalcOrder(int op,double &openLots,double &stopLossPrice,double &takeProfitPrice,datetime &expire) {
   if (op!=OP_BUY && op!=OP_SELL) return false;
   if (op==OP_BUY) {
      if (takeProfit!=0) takeProfitPrice=NormalizeDouble(Ask+takeProfit*_Point,_Digits);
      if (stopLoss!=0) stopLossPrice=NormalizeDouble(Bid-stopLoss*_Point,_Digits);
   } else if (op==OP_SELL) {
      if (takeProfit!=0) takeProfitPrice=NormalizeDouble(Bid-takeProfit*_Point,_Digits);
      if (stopLoss!=0) stopLossPrice=NormalizeDouble(Ask+stopLoss*_Point,_Digits);
   }
   if (lots==0) openLots=MarketInfo(_Symbol,MODE_MINLOT);
   else {
      double minLot;
      double maxLot;
      double lotStep;
      minLot=MarketInfo(_Symbol,MODE_MINLOT);
      maxLot=MarketInfo(_Symbol,MODE_MAXLOT);
      lotStep=MarketInfo(_Symbol,MODE_LOTSTEP);
      openLots=minLot+MathFloor((lots-minLot)/lotStep)*lotStep; 
      if (openLots<minLot) {
         openLots=minLot;
      } else if (openLots>maxLot) {
         openLots=maxLot;
      }
   }
   if (ttl!=0)
      expire=TimeCurrent()+ttl*PeriodSeconds(tf);
   else 
      expire=0;
   return true;
}
int EA::TradeByMarket(int op) {
   int ticket;
   double sl,tp;
   double tradeLots;
   int attempt;
   int err;
   datetime expire;
   ticket=-1;
   for(attempt=0;attempt<5;attempt++) {
      if (!AllowedOpen(op)) return -1;
      if ( IsTradeContextBusy() ) {
         Sleep(100);
         continue;
      }
      if (!CalcOrder(op,tradeLots,sl,tp,expire)) break;
      ResetLastError();
      if (AccountFreeMarginCheck(_Symbol,op,tradeLots)<=0 || GetLastError()==134) {
         return -1;
      }
      if (!IsConnected()) {
         Sleep(100);
         continue;
      }
      ticket=OrderSend(_Symbol,op,tradeLots,(op==OP_BUY?Ask:Bid),slippage,sl,tp,(expire==0?"":TimeToString(expire)),magik);
      if (ticket>=0) break;   // ok !
      err=GetLastError();
      if (err==ERR_BROKER_BUSY) {
         Sleep(100);
      } else if (err==ERR_REQUOTE || err==ERR_INVALID_PRICE) {
         if (attempt!=0) Sleep(100);
         RefreshRates();
      } else continue;
   }
   if (ticket<0) {
      OnError("Unable to "+(op==OP_BUY?"BUY":"SELL") );
   } else if (OrderSelect(ticket,SELECT_BY_TICKET,MODE_TRADES)) {
      if (panel!=NULL) {
         string s;
         //  : BUY|SELL #ticket price lots OK
         s=StringFormat("%s #%d %s %s OK",(OrderType()==OP_BUY?"BUY":"SELL"),OrderTicket(),DoubleToStr(OrderOpenPrice(),_Digits),DoubleToStr(OrderLots(),2) );
         panel.StateText(s,clrGreen,30);
      }
   }
   return ticket;
}

int EA::Buy() {
   int tt;
   if (panel!=NULL) panel.BeginBuy();
   tt=TradeByMarket(OP_BUY);
   if (panel!=NULL) panel.EndBuy();
   return tt;
}
int EA::Sell() {
   int tt;
   if (panel!=NULL) panel.BeginSell();
   tt=TradeByMarket(OP_SELL);
   if (panel!=NULL) panel.EndSell();
   return tt;
}

double EA::CloseAll(int op=-1) {
   int total;
   int pos;
   double profit=0.0;
   if (panel!=NULL) panel.BeginCloseAll();
   total=OrdersTotal();
   for(pos=total-1;pos>=0;pos--) {
      if (!AllowedClose()) return 0;
      if (!OrderSelect(pos,SELECT_BY_POS,MODE_TRADES)) continue;
      if (OrderSymbol()!=_Symbol || OrderMagicNumber()!=magik) continue;
      if (op==-1 || OrderType()==op) {
         profit+=CloseSelected();
      }       
   }
   if (panel!=NULL) panel.EndCloseAll();
   return profit;
}
double EA::CloseLast(int op=-1) {
   double profit;
   int total;
   int pos;
   int ticket=-1;
   datetime ticketTime=0;
   if (panel!=NULL) panel.BeginCloseLast();
   profit=0.0;
   total=OrdersTotal();
   //     
   for(pos=total-1;pos>=0;pos--) {
      if (!AllowedClose()) return 0;
      if (!OrderSelect(pos,SELECT_BY_POS,MODE_TRADES)) continue;
      if (OrderSymbol()!=_Symbol || OrderMagicNumber()!=magik) continue;
      if (OrderCloseTime()!=0) continue;
      if (op!=-1 && op!=OrderType()) continue;
      if (ticket==-1 || OrderOpenTime()>ticketTime || (OrderOpenTime()==ticketTime && ticket>OrderTicket()) ) {
         ticket=OrderTicket();
         ticketTime=OrderOpenTime();
      }
   }
   if (ticket!=-1 && OrderSelect(ticket,SELECT_BY_TICKET)) {
      return CloseSelected();
   }
   if (panel!=NULL) panel.EndCloseAll();
   return 0;
}
double EA::CloseExpired(int seconds,int op=-1) {
   double profit;
   int total;
   int pos;
   datetime expire;
   const int secInYear=365*24*60*60;
   profit=0.0;
   total=OrdersTotal();
   for(pos=total-1;pos>=0;pos--) {
      if (!AllowedClose()) return 0;
      if (!OrderSelect(pos,SELECT_BY_POS,MODE_TRADES)) continue;
      if (OrderSymbol()!=_Symbol || OrderMagicNumber()!=magik) continue;
      if (OrderCloseTime()!=0) continue;
      if (op!=-1 && op!=OrderType()) continue;
      if (seconds!=0 && OrderOpenTime()+seconds<=now) {
         profit+=CloseSelected();
      }
      ResetLastError();
      if (OrderComment()!="" && (expire=StrToTime(OrderComment()))<now && expire>now-secInYear && GetLastError()==ERR_NO_ERROR) {
         profit+=CloseSelected();
      }
   }
   return profit;      
}

double EA::CloseSelected() {
   double profit;
   double price;
   bool yes;
   int attempt;
   int err;
   int freeze;
   profit=0.0;
   price=0.0;
   yes=false;
   for(attempt=0;attempt<5;attempt++) {
      if (!IsConnected()) {
         Sleep(100);
         continue;
      }
      switch (OrderType()) {
         case OP_BUY: price=Bid;break;
         case OP_SELL: price=Ask;break;
      }
      freeze=(int)MarketInfo(_Symbol,MODE_FREEZELEVEL);
      if (freeze!=0 && price<=Ask+freeze*_Point && price>=Bid-freeze*_Point) {
         Sleep(100);
         continue;
      }
      yes=false;
      if (price!=0) {
         yes=OrderClose(OrderTicket(),OrderLots(),price,slippage);
      } else {
         yes=OrderDelete(OrderTicket());
      }
      if (yes) break;
      err=GetLastError();
      if (err==ERR_BROKER_BUSY) {
         Sleep(100);
      } else if (err==ERR_INVALID_PRICE || err==ERR_REQUOTE) {
         if (attempt) Sleep(100);
         RefreshRates();
      } else break;
   }
   if (yes) {
      profit=OrderProfit();
   } else {
      OnError("unable to close "+IntegerToString(OrderTicket()));
   }
   return profit;
}
void EA::Update() {
   int pos,total;
   datetime lastTime;
   int lastTicket;
   double lastProfit,totalProfit;
   
   if (panel==NULL) return;
   total=OrdersTotal();
   lastTicket=-1;
   lastTime=0;
   lastProfit=totalProfit=0.0;
   for(pos=total-1;pos>=0;pos--) {
      if (!OrderSelect(pos,SELECT_BY_POS,MODE_TRADES)) continue;
      if (OrderSymbol()!=_Symbol || OrderMagicNumber()!=magik) continue;
      if (OrderCloseTime()!=0) continue;
      if (OrderType()!=OP_BUY && OrderType()!=OP_SELL) continue;
      if (lastTime<OrderOpenTime() || (lastTime==OrderOpenTime() && lastTicket<OrderTicket()) ) {
         lastTime=OrderOpenTime();
         lastTicket=OrderTicket();
         lastProfit=OrderProfit();
      }
      totalProfit+=OrderProfit();
   }
   panel.UpdateTotals(totalProfit,lastProfit);
}
void EA::RequestBuy(void) {
   if (!IsTesting()) DrawSignal(OP_BUY,Time[0],Open[0]);
   if (mode==0) {
      if (IsTradeTime()) Buy();
      return;
   }
   request=REQUEST_BUY;
   requestTimeout=TimeCurrent()+90;
   if (panel!=NULL) panel.BeginRequest(request,requestTimeout,IsTradeTime());
}
void EA::RequestSell(void) {
   if (!IsTesting()) DrawSignal(OP_SELL,Time[0],Open[0]);
   if (mode==0) {
      if (IsTradeTime()) Sell();
      return;
   }
   request=REQUEST_SELL;
   requestTimeout=TimeCurrent()+90;
   if (panel!=NULL) panel.BeginRequest(request,requestTimeout,IsTradeTime());
}
void EA::RequestCloseAll(void) {
   if (mode==0) {
      CloseAll();
      return;
   }
   request=REQUEST_CLOSE_ALL;
   requestTimeout=TimeCurrent()+90;
   if (panel!=NULL) panel.BeginRequest(request,requestTimeout);
}
void EA::RequestCloseLast(void) {
   if (mode==0) {
      CloseLast();
      return;
   }
   request=REQUEST_CLOSE_LAST;
   requestTimeout=TimeCurrent()+90;
   if (panel!=NULL) panel.BeginRequest(request,requestTimeout);
}

void EA::DrawSignal(int op,datetime time,double price) {
   string obj;
   obj=name+".sig"+IntegerToString(time);
   if (!ObjectCreate(0,obj,OBJ_ARROW,0,time,price) && ObjectType(obj)!=OBJ_ARROW) return;
   ObjectSetInteger(0,obj,OBJPROP_TIME,time);
   ObjectSetDouble(0,obj,OBJPROP_PRICE,price);
   if (IsTradeTime(time)) {
      ObjectSetInteger(0,obj,OBJPROP_ARROWCODE,108);
   } else {
      ObjectSetInteger(0,obj,OBJPROP_ARROWCODE,163);
   }
   if (op==OP_BUY) {
      ObjectSetInteger(0,obj,OBJPROP_COLOR,clrBlue);
   } else {
      ObjectSetInteger(0,obj,OBJPROP_COLOR,clrRed);
   }
}